<?php

declare(strict_types = 1);

namespace Drupal\smart_migrate_cli\Commands;

use Consolidation\OutputFormatters\StructuredData\PropertyList;
use Consolidation\OutputFormatters\StructuredData\RenderCellCollectionTrait;
use Consolidation\SiteAlias\SiteAliasManagerAwareInterface;
use Consolidation\SiteAlias\SiteAliasManagerAwareTrait;
use Drupal\Component\Utility\Random;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Session\UserSession;
use Drupal\Core\Site\Settings;
use Drupal\Core\State\StateInterface;
use Drupal\migrate\Plugin\migrate\id_map\Sql;
use Drupal\migrate\Plugin\MigrateIdMapInterface;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\smart_migrate_cli\SmartMigrateCli;
use Drupal\smart_migrate_cli\Thread\MigrationThreadInterface;
use Drupal\smart_migrate_cli\SimpleMigrationManager;
use Drupal\smart_migrate_cli\Traits\SmartMigrateHelperTrait;
use Drupal\smart_migrate_cli\Traits\SmartMigrationConfigurationTrait;
use Drush\Commands\DrushCommands;
use Drush\Exceptions\UserAbortException;
use Drush\Utils\StringUtils;
use Symfony\Component\Console\Command\LockableTrait;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Output\ConsoleOutput;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Filesystem\Path;

/**
 * Smart Migrate CLI's own Drush commands.
 */
class SmartMigrateCliCommands extends DrushCommands implements SiteAliasManagerAwareInterface {

  use SiteAliasManagerAwareTrait;
  use SmartMigrateHelperTrait;
  use SmartMigrationConfigurationTrait;
  use LockableTrait;
  use RenderCellCollectionTrait;

  /**
   * Exit code used of processes which migration's requirements aren't met.
   *
   * @const int
   */
  const EXIT_REQUIREMENTS_NOT_MET = 10;

  /**
   * Exit code of single processes which migration threw an exception.
   *
   * @const int
   */
  const EXIT_PROCESS_CRITICAL = 100;

  /**
   * Requirements cache key for only required migrations.
   *
   * @const string
   */
  const REQUIREMENTS_KEY_REQUIRED_ONLY = 'required_only';

  /**
   * Requirements cache key for required and available optional migrations.
   *
   * @const string
   */
  const REQUIREMENTS_KEY_WITH_OPTIONALS = 'with_optionals';

  /**
   * The logger section used by SmartMigrateHelper's getNextUnblockedMigration.
   *
   * Yepp, bad design.
   *
   * @var \Symfony\Component\Console\Output\OutputInterface|null
   */
  private ?OutputInterface $loggerSection;

  /**
   * Storage for empty "placeholder" sections around progress bars.
   *
   * @var \Symfony\Component\Console\Output\OutputInterface[]
   */
  private array $barPlaceholderSections = [];

  /**
   * Error messages recorded during multi-thread migration.
   *
   * First level keys are migration plugin IDs, second level keys are md5 hashes
   * of the messages. Each value is an array where the "message" key contains
   * the message and the "count" key's value is the number of how many times the
   * error was triggered during migration of the specified plugin instance.
   *
   * @var array[][]
   */
  private array $errorOutputs = [];

  /**
   * The progress bars and their metadata.
   *
   * @var array[]
   *
   * Each "subthread" has a progress bar.
   *
   * @todo Provide our own object for these things.
   */
  private array $progressBars = [];

  /**
   * The threads.
   *
   * @var \Drupal\smart_migrate_cli\Thread\MigrationThreadInterface[]
   */
  private array $threads = [];

  /**
   * List of the migration plugin instances to execute.
   *
   * @var \Drupal\migrate\Plugin\MigrationInterface[]
   */
  private array $list = [];

  /**
   * List of the migration plugin IDs which should be executed.
   *
   * @var string[]
   */
  private array $migrationsToProcess = [];

  /**
   * List of the already processed migrations' plugin IDs.
   *
   * @var string[]
   */
  private array $migrationsProcessed = [];

  /**
   * List of the fully processed migrations' plugin IDs.
   *
   * @var string[]
   */
  private array $migrationsFullyProcessed = [];

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected StateInterface $state;

  /**
   * Account switcher.
   *
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected AccountSwitcherInterface $accountSwitcher;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Light-weight migration plugin manager service optimized for multi-thread.
   *
   * @var \Drupal\smart_migrate_cli\SimpleMigrationManager|null
   */
  protected ?SimpleMigrationManager $migrationPluginManager;

  /**
   * Light-weight migration plugin manager service optimized for multi-thread.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected DateFormatterInterface $dateFormatter;

  /**
   * The active database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Section of the total progress feedback.
   *
   * @var \Symfony\Component\Console\Output\OutputInterface|null
   */
  protected $totalProgressSection;

  /**
   * The total progress feedback.
   *
   * @var \Drupal\smart_migrate_cli\TotalProgress
   */
  protected $totalProgress;

  /**
   * Constructs a new SmartMigrateCliCommands instance.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity_type manager.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter.
   * @param \Drupal\Core\Database\Connection $connection
   *   The active database connection.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\smart_migrate_cli\SimpleMigrationManager|null $simple_migration_plugin_manager
   *   The light-weight migration plugin manager.
   */
  public function __construct(StateInterface $state, AccountSwitcherInterface $account_switcher, EntityTypeManagerInterface $entity_type_manager, DateFormatterInterface $date_formatter, Connection $connection, ModuleHandlerInterface $module_handler, ?SimpleMigrationManager $simple_migration_plugin_manager = NULL) {
    parent::__construct();
    $this->state = $state;
    $this->accountSwitcher = $account_switcher;
    $this->entityTypeManager = $entity_type_manager;
    $this->migrationPluginManager = $simple_migration_plugin_manager;
    $this->dateFormatter = $date_formatter;
    $this->database = $connection;
    $this->moduleHandler = $module_handler;
    $this->loggerSection = $this->output();
  }

  /**
   * Executes migrations in the optimal order, in multiple threads.
   *
   * @param array|null $options
   *   Command line options.
   *
   * @command smart-migrate:multi-thread-import
   *
   * @option tag
   *   Tag of the migrations to execute.
   * @option threads
   *   Maximal number of threads to use. Optional, defaults to 4.
   * @option memory
   *   Display memory usage even if verbosity is normal.
   * @option estimate
   *   Display estimated time even if verbosity is normal.
   * @option run-at-the-end
   *   Execute the specified migrations at the end of the process.
   *
   * @validate-module-enabled migrate
   *
   * @aliases smimt,smmti
   *
   * @throws \Exception
   *   When not enough options were provided or no migration was found.
   */
  public function mtImport(?array $options = [
    'tag' => self::REQ,
    'threads' => self::OPT,
    'memory' => self::OPT,
    'estimate' => self::OPT,
    'run-at-the-end' => self::OPT,
  ]): int {
    return $this->multiThreadProcess($options, 'import');
  }

  /**
   * Executes rollback-removed operation in multiple threads.
   *
   * @param array|null $options
   *   Command line options.
   *
   * @command smart-migrate:multi-thread-rollback-removed
   *
   * @option tag
   *   Tag of the migrations to execute.
   * @option threads
   *   Maximal number of threads to use. Optional, defaults to 4.
   * @option memory
   *   Display memory usage even if verbosity is normal.
   * @option estimate
   *   Display estimated time even if verbosity is normal.
   * @option run-at-the-end
   *   Execute the specified migrations at the end of the process.
   *
   * @validate-module-enabled migrate
   *
   * @aliases smrdmt,smmtrd
   *
   * @throws \Exception
   *   When not enough options were provided or no migration was found.
   */
  public function mtRollbackRemoved(?array $options = [
    'tag' => self::REQ,
    'threads' => self::OPT,
    'memory' => self::OPT,
    'estimate' => self::OPT,
    'run-at-the-end' => self::OPT,
  ]): int {
    return $this->multiThreadProcess($options, 'rollback-removed');
  }

  /**
   * Performs the given migration import operation in multiple threads.
   *
   * @param array $options
   *   Options of the initiator command.
   * @param string $operation
   *   The migration operation to perform.
   *
   * @return int
   *   The exit code. Might be self::EXIT_SUCCESS or self::EXIT_FAILURE.
   */
  protected function multiThreadProcess(array $options, string $operation): int {
    if (!$this->lock(SmartMigrateCli::MULTI_THREAD_ID)) {
      $this->logger()->error('A multi-thread command is already running in another process.');
      return self::EXIT_FAILURE;
    }
    $start = time();

    if ($return = $this->initializeMultiThreadCommand($options, $operation)) {
      $this->release();
      return $return;
    };

    // Register our SIGINT and SIGTERM signals. If the main process gets one of
    // these, we have to stop all the subprocesses too.
    if (extension_loaded('pcntl')) {
      pcntl_signal(SIGTERM, static::stopSignalClosure());
      pcntl_signal(SIGINT, static::stopSignalClosure());
      pcntl_async_signals(TRUE);
    }

    while (!(empty($this->getBusyThreads()) && empty($this->migrationsToProcess))) {
      try {
        while (count($busy_threads = $this->getBusyThreads()) < count($this->threads) && !empty($this->migrationsToProcess)) {
          if (!($migration_id = $this->getNextUnblockedMigration($operation))) {
            if (empty($busy_threads)) {
              // There is no unblocked migration, and we don't have any active
              // processes. This means we cannot continue.
              break 2;
            }
            // Let's wait until a process finishes.
            // @todo Move SmartMigrateHelperTrait::getNextUnblockedMigration's
            //   logging here.
            break 1;
          }
          $this->scheduleMigrationProcess($migration_id);
        }

        $this->handleThreadProgress();
        usleep((int) (SmartMigrateCli::POLL_WAIT_SECONDS * 1000 * 1000));
      }
      catch (\Exception $e) {
        if (static::isStopSignal($e)) {
          // Do not report temporary error messages.
          $this->errorOutputs = array_intersect_key($this->errorOutputs, $this->migrationsProcessed);

          // Kill threads' subprocesses.
          foreach ($this->activeThreads() as $thread) {
            try {
              $thread->stop();
            }
            catch (\Throwable $t) {
            }
          }

          // Do not report "unmet" migration dependencies.
          $this->migrationsToProcess = [];
          break 1;
        }
        throw $e;
      }
    }

    return $this->finishMultiThreadCommand($start, !empty($e));
  }

  /**
   * Executes one migration.
   *
   * @command smart-migrate:single-import
   *
   * @usage smart-migrate:single-import foo
   *   Executes "foo" migration.
   *
   * @validate-module-enabled migrate
   *
   * @hidden
   */
  public function singleImport(): int {
    $migration_id = getenv(SmartMigrateCli::MIGRATION_PLUGIN_ID_ENV);
    $migrations = $migration_id ? $this->migrationPluginManager->createInstances([$migration_id]) : [];
    if (count($migrations) !== 1) {
      $this->logger()->error('Too many, or no migrations found.');
      $return = self::EXIT_FAILURE;
    }
    $migration = reset($migrations);
    $message_to_parent = ['id' => $migration_id];
    $verbosity = $this->output()->getVerbosity();
    if (
      !isset($return) &&
      ($requirement_message = $this->getMigrationRequirementsMessages($migration))
    ) {
      $this->logger()->error($requirement_message);
      $return = static::EXIT_REQUIREMENTS_NOT_MET;
    }

    if (empty($return)) {
      // Switch to user 1. This is what we usually do on the UI as well.
      // The user ID mostly affects the user reference (base) fields of migrated
      // entities which don't have source data - e.g. the owner of user or
      // taxonomy term translations.
      $this->accountSwitcher->switchTo(new UserSession(['uid' => 1]));
      try {
        $follow_up_ids = $this->executeMigration($migration);
        if (!empty($follow_up_ids)) {
          $message_to_parent['follow_ups'] = $follow_up_ids;
        }
        $return = self::EXIT_SUCCESS;
      }
      catch (\Exception $e) {
        $this->logger()->critical($e->getMessage());
        $return = self::EXIT_PROCESS_CRITICAL;
      }
      $this->accountSwitcher->switchBack();
    }

    $this->output()->setVerbosity(OutputInterface::VERBOSITY_NORMAL);
    $this->output()->writeln(json_encode($message_to_parent));
    $this->output()->setVerbosity($verbosity);

    return $return;
  }

  /**
   * Rolls back items removed from source since the last import.
   *
   * @command smart-migrate:single-rollback-removed
   *
   * @usage smart-migrate:single-rollback-removed foo
   *   Rolls back items in "foo" migration which were deleted from source.
   *
   * @validate-module-enabled migrate
   *
   * @hidden
   */
  public function singleRollbackRemoved(): int {
    $migration_id = getenv(SmartMigrateCli::MIGRATION_PLUGIN_ID_ENV);
    $migrations = $migration_id ? $this->migrationPluginManager->createInstances([$migration_id]) : [];
    if (count($migrations) !== 1) {
      $this->logger()->error('Too many, or no migrations found.');
      $return = self::EXIT_FAILURE;
    }
    $migration = reset($migrations);
    $message_to_parent = ['id' => $migration_id];
    $verbosity = $this->output()->getVerbosity();
    if (
      !isset($return) &&
      ($requirement_message = $this->getMigrationRequirementsMessages($migration))
    ) {
      $this->logger()->error($requirement_message);
      $return = static::EXIT_REQUIREMENTS_NOT_MET;
    }

    if (empty($return)) {
      // Switch to user 1. This is what we usually do on the UI as well.
      // The user ID mostly affects the user reference (base) fields of migrated
      // entities which don't have source data - e.g. the owner of user or
      // taxonomy term translations.
      $this->accountSwitcher->switchTo(new UserSession(['uid' => 1]));
      try {
        $this->executeMigration($migration, 'rollback-removed');
        $return = self::EXIT_SUCCESS;
      }
      catch (\Exception $e) {
        $this->logger()->critical($e->getMessage());
        $return = self::EXIT_PROCESS_CRITICAL;
      }
      $this->accountSwitcher->switchBack();
    }

    $this->output()->setVerbosity(OutputInterface::VERBOSITY_NORMAL);
    $this->output()->writeln(json_encode($message_to_parent));
    $this->output()->setVerbosity($verbosity);

    return $return;
  }

  /**
   * Returns the active threads.
   *
   * @return \Drupal\smart_migrate_cli\Thread\MigrationThreadInterface[]
   *   The active threads.
   */
  protected function activeThreads(): array {
    return array_filter(
      $this->threads,
      function (MigrationThreadInterface $thread) {
        return $thread->isActive();
      }
    );
  }

  /**
   * Initializes multi-thread process task.
   *
   * @param array $options
   *   Options of the initiator command.
   * @param string $operation
   *   The migration operation to initialize. Defaults to "import".
   *
   * @return int
   *   The exit code. Might be self::EXIT_SUCCESS or self::EXIT_FAILURE.
   */
  protected function initializeMultiThreadCommand(array $options, string $operation = 'import'): int {
    // Initialize our custom progress bar format.
    $format_template = $format_template_nomax = SmartMigrateCli::PROGRESS_BAR_FORMAT;
    $format_template_total = $format_template_total_nomax = SmartMigrateCli::PROGRESS_BAR_FORMAT_TOTAL;
    if (
      $this->output()->getVerbosity() > OutputInterface::VERBOSITY_NORMAL ||
      $options['estimate']
    ) {

      $format_template_total .= ' | %remaining:7s%';
      $format_template .= ' | %remaining:7s%';
    }
    if (
      $this->output()->getVerbosity() > OutputInterface::VERBOSITY_NORMAL ||
      $options['memory']
    ) {
      $format_template_total .= ' | %memory:6s%';
      $format_template_total_nomax .= ' | %memory:6s%';
    }
    ProgressBar::setFormatDefinition(SmartMigrateCli::PROGRESS_BAR_FORMAT_ID, $format_template);
    ProgressBar::setFormatDefinition(SmartMigrateCli::PROGRESS_BAR_FORMAT_ID_NOMAX, $format_template_nomax);
    ProgressBar::setFormatDefinition(SmartMigrateCli::PROGRESS_BAR_FORMAT_ID_TOTAL, $format_template_total);
    ProgressBar::setFormatDefinition(SmartMigrateCli::PROGRESS_BAR_FORMAT_ID_TOTAL_NOMAX, $format_template_total_nomax);

    $this->output()->writeln('Getting migration list...');
    $list = $this->getMigrationList($this->input()->getOption('tag'));
    if (empty($list) || empty($list = reset($list))) {
      $this->logger()->error(sprintf(
        "No migrations found with the given tag %s.",
        $this->input()->getOption('tag')
      ));
      return self::EXIT_FAILURE;
    }
    $run_at_the_end = $options['run-at-the-end']
      ? StringUtils::csvToArray($options['run-at-the-end'])
      : array_keys(
        array_filter(
          $list,
          function (MigrationInterface $migration) {
            return in_array(
              $migration->getDestinationConfiguration()['plugin'],
              ['entity:path_alias', 'entity:menu_link_content'],
              TRUE
            );
          }
        )
      );
    $this->initializeMultiThreadProcessing($list, $run_at_the_end, $format_template_total, $operation);

    // Determine how many threads we can use.
    $threads_input = $this->input()->getOption('threads');
    $threads = is_numeric($threads_input) ? (int) $threads_input : 4;
    $this->output()->writeln(sprintf(
      "Starting migration in %s threads",
      $threads
    ));

    // Initialize UI.
    $this->initializeThreads($threads, $operation);

    $this->loggerSection = $this->output()->section();
    $this->setLoggerOutput($this->loggerSection);

    return self::EXIT_SUCCESS;
  }

  /**
   * Finishes multi-thread process task.
   */
  protected function finishMultiThreadCommand(int $process_start_timestamp, bool $stopped): int {
    foreach ($this->threads as $thread) {
      $thread->stop();
    }

    $this->totalProgress->finish();
    foreach ($this->barPlaceholderSections as $section) {
      $section->clear();
    }
    $elapsed_time = $this->dateFormatter->formatDiff(
      $process_start_timestamp,
      time(),
      ['granularity' => 3, 'langcode' => 'en']
    );
    $this->finalReport();

    if ($stopped) {
      $this->logger()->warning('Migration process has been stopped.');
      return self::EXIT_FAILURE;
    }

    if (!empty($this->migrationsToProcess)) {
      $this->logger()->error('Cannot properly finish import process because the following migrations have unmet (required) dependencies: ' . implode(', ', $this->migrationsToProcess));
      return self::EXIT_FAILURE;
    }

    $this->loggerSection->clear();
    $this->logger()->success(sprintf(
      'Migration process has been finished. It took %s.',
      $elapsed_time
    ));
    $this->release();
    return self::EXIT_SUCCESS;
  }

  /**
   * Returns the sorted and filtered list of migrations with the given tag.
   *
   * @param string $tag
   *   The migration tag of the migrations.
   *
   * @return \Drupal\migrate\Plugin\MigrationInterface[][]
   *   List of the discovered and executable migrations.
   */
  protected function getMigrationList(string $tag): array {
    $grouped_list = $this->doGetGroupedMigrationList(NULL, $tag);
    $grouped_list = $this->filterMigrationList($grouped_list);
    return $this->orderMigrationList($grouped_list, TRUE);
  }

  /**
   * Changes the logger output to the specified output.
   *
   * @param \Symfony\Component\Console\Output\OutputInterface|null $output
   *   The output to use by the logger.
   */
  private function setLoggerOutput(OutputInterface $output = NULL): void {
    $this->logger()->setErrorStream($output ?? $this->output());
    $this->logger()->setOutputStream($output ?? $this->output());
  }

  /**
   * Adds migrate source data to settings.php.
   *
   * @command smart-migrate:configure
   *
   * @option db-url
   *   Url of database connection. Format:
   *   "driver://username:password@host:port/$database#table_prefix".
   * @option source-base-path
   *   Absolute path to the source Drupal instance's web root. Necessary for
   *   migrating files. The path of the files will be calculated with also
   *   evaluating the files_public_path variable of the source Drupal instance,
   *   so please do not add "sites/default/files" to the end of this path.
   * @option source-files-private
   *   Absolute path to the source Drupal instance's private file directory.
   * @option sites-subdir
   *   Name of directory under <info>sites</info> which should be configured.
   *   Optional, defaults to the 'default' directory.
   * @option connection
   *   Name of the database connection to use. Defaults to <info>migrate</info>.
   *
   * @usage smart-migrate:configure --db-url=mysql://user:password@127.0.0.1/foo
   *   Adds the 'foo' mysql database connection as migration source to the
   *   settings.php.
   * @usage smart-migrate:configure --source-base-path=/web/drupal-7
   *   Sets the '/web/drupal-7' directory as migration source webroot.
   * @usage smart-migrate:configure --source-files-private=/web/private
   *   Sets the '/web/private' directory as migration source path of private
   *   files.
   * @usage smart-migrate:configure --db-url=sqlite/localhost/test.sqlite --source-base-path=/web/drupal-7 --source-files-private=/web/private
   *   Multiple options could set in one command.
   *
   * @aliases smconf,sm-conf
   */
  public function migrationConfigure(array $options = [
    'db-url' => self::REQ,
    'source-base-path' => self::REQ,
    'source-files-private' => self::REQ,
    'sites-subdir' => self::OPT,
    'connection' => 'migrate',
  ]): void {
    $db_url = $this->input()->getOption('db-url');
    $files_source = $this->input()->getOption('source-base-path');
    $files_private_source = $this->input()->getOption('source-files-private');

    // If all options are empty, throw an exception.
    if (!$db_url && !$files_source && !$files_private_source) {
      throw new \LogicException('To configure any value for migration, one of --db-url, --source-base-path or --source-files-private options must be defined.');
    }

    // Handle aliases, if any.
    // @see SiteInstallCommands::pre().
    $alias_record = $this->siteAliasManager()->getSelf();
    $root = $alias_record->root();
    if (!($dir = $this->input()->getOption('sites-subdir'))) {
      $dir = $this->getSitesSubdirFromUri($root, $alias_record->get('uri'));
    }

    if (!$dir) {
      throw new \Exception('Could not determine target sites directory for site to install. Use --sites-subdir to specify.');
    }

    $sites_subdir = Path::join('sites', $dir);
    $settingsfile = Path::join($sites_subdir, 'settings.php');
    $settingsfile_write = Path::join($root, $settingsfile);

    if ($db_url) {
      $key = $this->input()->getOption('connection');
      $connection_info = Database::convertDbUrlToConnectionInfo($db_url, $root);

      // Let's test the connection before writing it to settings.php.
      try {
        // Find a connection key that doesn't exist.
        $random = new Random();
        do {
          $test_key = 'tmp' . strtolower($random->name(6));
        } while (Database::getConnectionInfo($test_key));
        Database::addConnectionInfo($test_key, 'default', $connection_info);
        Database::getConnection('default', $test_key)->schema()->findTables('%');
        // Restore original connection structure.
        Database::removeConnection($test_key);
      }
      catch (\Exception $e) {
        throw new \Exception(sprintf(
          "Cannot connect to the database given in DB url '%s'. Maybe a typo?",
          $db_url
        ));
      }

      if (
        Database::getConnectionInfo($key) &&
        $db_url !== $this->getDatabaseUrl(Database::getConnection('default', $key)) &&
        !$this->io()->confirm(dt('The given database connection already exists. Do you really want to override?'))
      ) {
        throw new UserAbortException();
      }

      if (
        !Database::getConnectionInfo($key) ||
        $db_url !== $this->getDatabaseUrl(Database::getConnection('default', $key))
      ) {
        $settings['databases'][$key]['default'] = (object) [
          'value' => $connection_info,
          'required' => TRUE,
        ];
      }
    }

    if ($files_source) {
      $current_base_path = Settings::get('source_base_path');
      if (
        $current_base_path &&
        $current_base_path !== $files_source &&
        !$this->io()->confirm(dt(
          "The source base path is already configured, and it is set to '!value'. Do you really want to override?",
          ['!value' => $current_base_path]
        ))
      ) {
        throw new UserAbortException();
      }
      $settings['settings']['source_base_path'] = (object) [
        'value' => $files_source,
        'required' => TRUE,
      ];
    }

    if ($files_private_source) {
      $current_private_source = Settings::get('source_private_file_path');
      if (
        $current_private_source &&
        $current_private_source !== $files_private_source &&
        !$this->io()->confirm(dt(
          "The source private files path is already configured, and it is set to '!value'. Do you really want to override?",
          ['!value' => $current_private_source]
        ))
      ) {
        throw new UserAbortException();
      }
      $settings['settings']['source_private_file_path'] = (object) [
        'value' => $files_private_source,
        'required' => TRUE,
      ];
    }

    if (!empty($settings)) {
      if (!is_writable($settingsfile_write) && !chmod($settingsfile_write, 0644)) {
        throw new \Exception(sprintf(
          "Cannot make '%s' file writable. Please modify file permissions manually and re-execute the command!",
          $settingsfile_write
        ));
      }

      include_once $root . '/core/includes/install.inc';
      drupal_rewrite_settings($settings, $settingsfile_write);
    }

    if ($db_url) {
      $this->state->set('smart_migrate_cli.source_connection', [
        'key' => $key,
        'target' => 'default',
      ]);
      $this->state->set('migrate.fallback_state_key', 'smart_migrate_cli.source_connection');
    }

    $this->logger()->success('Migration configuration was written to settings.php.');
  }

  /**
   * Determine an appropriate site subdir name to use for the provided URI.
   *
   * @param string $root
   *   The Drupal root.
   * @param string $uri
   *   The site URI.
   *
   * @return string|null
   *   The name of the directory of the specified Drupal site, or NULL if it
   *   cannot be determined.
   */
  protected function getSitesSubdirFromUri(string $root, string $uri): ?string {
    $dir = strtolower($uri);
    // Always accept simple uris (e.g. 'dev', 'stage', etc).
    if (preg_match('#^[a-z0-9_-]*$#', $dir)) {
      return $dir;
    }
    // Strip off the protocol from the provided uri -- however,
    // now we will require that the sites subdir already exist.
    $dir = preg_replace('#[^/]*/*#', '', $dir);
    if ($dir && file_exists(Path::join($root, $dir))) {
      return $dir;
    }
    // Find the dir from sites.php file.
    $sites_file = $root . '/sites/sites.php';
    if (file_exists($sites_file)) {
      $sites = [];
      include $sites_file;
      if (!empty($sites) && array_key_exists($uri, $sites)) {
        return $sites[$uri];
      }
    }
    // Fall back to default directory if it exists.
    if (file_exists(Path::join($root, 'sites', 'default'))) {
      return 'default';
    }
    return NULL;
  }

  /**
   * Returns the database URL of the given connection.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The connection.
   *
   * @return string
   *   The database URL representation of the given connection.
   */
  protected function getDatabaseUrl(Connection $connection): string {
    $connection_info = $connection->getConnectionOptions();
    $driver = $connection_info['driver'];
    $user = $connection_info['username'] ?? NULL;
    $pass = $connection_info['password'] ?? NULL;
    $prefix = $connection_info['prefix'];
    $host = strpos($connection_info['database'], DRUPAL_ROOT) === 0
      ? 'localhost'
      : $connection_info['host'] ?? NULL;
    $port = $connection_info['port'] ?? NULL;
    $db_name = strpos($connection_info['database'], DRUPAL_ROOT) === 0
      ? substr($connection_info['database'], strlen(DRUPAL_ROOT) + 1)
      : $connection_info['database'];
    $server = implode(
      '@',
      array_filter([
        implode(':', array_filter([$user, $pass])),
        implode(':', array_filter([$host, $port])),
      ])
    );
    $table_and_prefix = implode('#', array_filter([
      $db_name,
      $prefix,
    ]));
    return "$driver://$server/$table_and_prefix";
  }

  /**
   * Closure called when operation is aborted.
   *
   * @return \Closure
   *   Closure called when operation is aborted.
   */
  protected static function stopSignalClosure(): \Closure {
    return function () {
      throw new \Exception(SmartMigrateCli::MULTI_THREAD_ID);
    };
  }

  /**
   * Determines whether the throwable is a stop signal.
   *
   * @param \Throwable $t
   *   A throwable to check.
   *
   * @return bool
   *   Whether the throwable is a stop signal.
   */
  protected static function isStopSignal(\Throwable $t): bool {
    return $t->getMessage() === SmartMigrateCli::MULTI_THREAD_ID;
  }

  /**
   * Resets all failed rows of the given migrations.
   *
   * @param array|null $options
   *   Command line options.
   *
   * @command smart-migrate:reset-failed
   *
   * @option tag
   *   Tag of the migrations to reset.
   *
   * @validate-module-enabled migrate
   *
   * @aliases smrf
   *
   * @throws \Exception
   *   When not enough options were provided or no migration was found.
   */
  public function migrateReset(?array $options = [
    'tag' => self::REQ,
  ]): int {
    $this->output()->writeln('Getting migration list...');
    $per_tag_list = $this->getMigrationList($this->input()->getOption('tag'));
    if (empty($per_tag_list) || empty($list = reset($per_tag_list))) {
      $this->logger()->error(sprintf(
        "No migrations found with the given tag %s.",
        $this->input()->getOption('tag')
      ));
      return self::EXIT_FAILURE;
    }

    $bar_section = $this->output() instanceof ConsoleOutput
      ? $this->output()->section()
      : $this->output();
    $bar = new ProgressBar($bar_section, count($list));
    $bar->start();

    foreach ($list as $instance) {
      assert($instance instanceof MigrationInterface);
      if (($id_map = $instance->getIdMap())->errorCount()) {
        if (
          $id_map instanceof Sql &&
          !$this->database->schema()->tableExists($id_map->messageTableName())
        ) {
          continue;
        }

        $id_map->rewind();
        while ($id_map->valid()) {
          $source_keys = $id_map->currentSource();
          $row = $id_map->getRowBySource($source_keys);
          if (!empty($row['source_row_status']) && $row['source_row_status'] == MigrateIdMapInterface::STATUS_FAILED) {
            $id_map->delete($source_keys);
          }
          $id_map->next();
        }

        $bar->advance();
      }
    }

    $bar->finish();
    $bar->clear();
    $this->output()->writeln('Failed rows were reset.');
    return static::EXIT_SUCCESS;
  }

  /**
   * An overview of the environment - Drush and Drupal.
   *
   * @command smart-migrate:status
   * @aliases sms
   * @table-style compact
   * @list-delimiter :
   * @field-labels
   *   php-memory: PHP memory
   *   host-db-uri: Host DB URI
   *   source-db-uri: Source DB URI
   *   host-isolation: Host isolation level
   *   source-isolation: Source isolation level
   *   joinability: Source and ID Map joinable
   * @default-fields php-memory,host-db-uri,source-db-uri,host-isolation,source-isolation,joinability
   * @pipe-format json
   */
  public function status(array $options = ['format' => 'table']): PropertyList {
    $result_data = [
      'php-memory' => ini_get('memory_limit'),
      'host-db-uri' => static::getConnectionInfoAsUrl($this->database->getKey(), $this->database->getTarget()),
      'host-isolation' => $this->getTransactionIsolationLevel($this->database->getKey(), $this->database->getTarget()),
    ];

    if ($this->moduleHandler->moduleExists('migrate')) {
      $migrate_source_state_key = $this->state->get('migrate.fallback_state_key');
      $migrate_source = $migrate_source_state_key
        ? $this->state->get($migrate_source_state_key)
        : [];
      $migrate_source += ['target' => 'default', 'key' => 'migrate'];
      if (
        !empty(Database::getConnectionInfo($migrate_source['key'])) &&
        !empty(Database::getConnectionInfo($migrate_source['key'])[$migrate_source['target']])
      ) {
        $source_connection_found = TRUE;
        $result_data['source-db-uri'] = static::getConnectionInfoAsUrl($migrate_source['key'], $migrate_source['target']);
        $result_data['source-isolation'] = $this->getTransactionIsolationLevel($migrate_source['key'], $migrate_source['target']);
      }

      if (!empty($source_connection_found) && $this->moduleHandler->moduleExists('migrate_drupal')) {
        $test_migration = $this->migrationPluginManager->createStubMigration([
          'id' => 'smc-test',
          'source' => ['plugin' => 'variable'],
          'destination' => ['plugin' => 'null'],
        ]);
        $var_source = $test_migration->getSourcePlugin();
        $reflection = new \ReflectionClass($var_source);
        $joinable = $reflection->getMethod('mapJoinable');
        $joinable->setAccessible(TRUE);
        $result_data['joinability'] = $joinable->invoke($var_source) ? '<info>Yes</info>' : '<error>No</error>';
      }
    }

    $result = new PropertyList($result_data);

    return $result;
  }

  /**
   * Returns the isolation level of the given DB connection.
   *
   * @param string $key
   *   (Optional) The database connection key.
   * @param string $target
   *   (Optional) The database connection target.
   *
   * @return string
   *   Isolation level of the given DB connection.
   */
  protected function getTransactionIsolationLevel(string $key = 'default', string $target = 'default'): string {
    $connection = Database::getConnection($target, $key);

    // The database variable "tx_isolation" has been removed in MySQL v8.0.3 and
    // has been replaced by "transaction_isolation".
    // @see https://dev.mysql.com/doc/refman/5.7/en/server-system-variables.html#sysvar_tx_isolation
    // @see https://dev.mysql.com/doc/refman/8.0/en/added-deprecated-removed.html
    $query = !$connection->isMariaDb() && version_compare($connection->version(), '8.0.2-AnyName', '>')
      ? 'SELECT @@SESSION.transaction_isolation'
      : 'SELECT @@SESSION.tx_isolation';

    try {
      return $connection->query($query)->fetchField() ?: '<comment>NaN</comment>';
    }
    catch (\Exception $e) {
    }

    return '<error>Fetching isolation triggered an exception</error>';
  }

  /**
   * Gets database connection info as a URL.
   *
   * @param string $key
   *   (Optional) The database connection key.
   * @param string $target
   *   (Optional) The database connection target.
   *
   * @return string
   *   The connection info as a URL.
   *
   * @throws \RuntimeException
   *   When the database connection is not defined.
   *
   * @see \Drupal\Core\Database\Database::getConnectionInfoAsUrl()
   */
  public static function getConnectionInfoAsUrl(string $key = 'default', string $target = 'default'): string {
    $db_info = Database::getConnectionInfo($key);
    if (empty($db_info) || empty($db_info[$target])) {
      throw new \RuntimeException(sprintf(
        "Database connection '%s' not defined or missing the '%s' settings",
        $key,
        $target
      ));
    }
    $namespace = $db_info[$target]['namespace'];

    // If the driver namespace is within a Drupal module, add the module name
    // to the connection options to make it easy for the connection class's
    // createUrlFromConnectionOptions() method to add it to the URL.
    if (static::isWithinModuleNamespace($namespace)) {
      $db_info[$target]['module'] = explode('\\', $namespace)[1];
    }

    $connection_class = $namespace . '\\Connection';
    return $connection_class::createUrlFromConnectionOptions($db_info[$target]);
  }

  /**
   * Checks whether a namespace is within the namespace of a Drupal module.
   *
   * This can be used to determine if a database driver's namespace is provided
   * by a Drupal module.
   *
   * @param string $namespace
   *   The namespace (for example, of a database driver) to check.
   *
   * @return bool
   *   TRUE if the passed in namespace is a sub-namespace of a Drupal module's
   *   namespace.
   *
   * @see \Drupal\Core\Database\Database::isWithinModuleNamespace()
   */
  private static function isWithinModuleNamespace(string $namespace): bool {
    [$first, $second] = explode('\\', $namespace, 3);

    // The namespace for Drupal modules is Drupal\MODULE_NAME, and the module
    // name must be all lowercase. Second-level namespaces containing uppercase
    // letters (e.g., "Core", "Component", "Driver") are not modules.
    // @see \Drupal\Core\DrupalKernel::getModuleNamespacesPsr4()
    // @see https://www.drupal.org/docs/8/creating-custom-modules/naming-and-placing-your-drupal-8-module#s-name-your-module
    return ($first === 'Drupal' && strtolower($second) === $second);
  }

}
