<?php

declare(strict_types = 1);

namespace Drupal\smart_migrate_cli;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\migrate\Exception\RequirementsException;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate\Plugin\MigrationPluginManager;
use Drupal\migrate\Plugin\MigrationPluginManagerInterface;
use Drupal\migrate\Plugin\RequirementsInterface;

/**
 * Lightweight migration plugin manager.
 */
class SimpleMigrationManager extends MigrationPluginManager {

  /**
   * The actual migration plugin manager.
   *
   * @var \Drupal\migrate\Plugin\MigrationPluginManagerInterface
   */
  protected $currentPluginManager;

  /**
   * Construct a migration plugin manager.
   *
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache backend for the definitions.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\migrate\Plugin\MigrationPluginManagerInterface $plugin_manager
   *   The migration plugin manager.
   */
  public function __construct(ModuleHandlerInterface $module_handler, CacheBackendInterface $cache_backend, LanguageManagerInterface $language_manager, MigrationPluginManagerInterface $plugin_manager) {
    parent::__construct($module_handler, $cache_backend, $language_manager);
    $this->currentPluginManager = $plugin_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function getDefinitions() {
    if (!isset($this->definitions)) {
      $this->definitions = $this->currentPluginManager->getDefinitions();
    }
    return $this->definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function clearCachedDefinitions() {
    $this->definitions = NULL;
    $this->currentPluginManager->clearCachedDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  public function buildDependencyMigration(array $migrations, array $dynamic_ids) {
    // If caller is the manager's createInstances method, we don't build the
    // dependency graph. Smart Migrate CLI will do this separately.
    $is_createinstances_call = array_reduce(
      debug_backtrace(0, 2),
      function (bool $carry, array $stack) {
        if ($carry) {
          return $carry;
        }
        $class = $stack['class'] ?? NULL;
        return $class &&
          in_array(MigrationPluginManagerInterface::class, class_implements($class)) &&
          ($stack['function'] ?? NULL) === 'createInstances';
      },
      FALSE
    );
    if ($is_createinstances_call) {
      return $migrations;
    }

    return is_callable([$this->currentPluginManager, 'buildDependencyMigration'])
      ? $this->currentPluginManager->buildDependencyMigration($migrations, $dynamic_ids)
      : parent::buildDependencyMigration($migrations, $dynamic_ids);
  }

  /**
   * Checks if requirements for the given migration plugin instance are met.
   *
   * This is an optimized version of the requirements check built into
   * migrations: this performs the same check, but operates on a clone, which is
   * removed from PHP memory after the check is done.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance to check.
   * @param bool $check_dependencies
   *   Whether dependencies also should be checked or not. Optional, defaults to
   *   TRUE.
   *
   * @return bool
   *   Whether requirements are met or not.
   */
  public function migrationRequirementsMet(MigrationInterface $migration, bool $check_dependencies = TRUE): bool {
    $clone = clone $migration;
    // Check whether the current migration source and destination plugin
    // requirements are met or not.
    try {
      if ($clone->getSourcePlugin() instanceof RequirementsInterface) {
        $clone->getSourcePlugin()->checkRequirements();
      }
      if ($clone->getDestinationPlugin() instanceof RequirementsInterface) {
        $clone->getDestinationPlugin()->checkRequirements();
      }

      if ($check_dependencies && !empty($requirements = $clone->getRequirements())) {
        // Requirements need to be checked.
        $required_migrations = $this->createInstances($requirements);

        if (empty(array_diff($requirements, array_keys($required_migrations)))) {
          // Check if the dependencies are in good shape.
          foreach ($required_migrations as $required_migration) {
            if (!$required_migration->allRowsProcessed()) {
              $return = FALSE;
              break 1;
            }
          }
        }
        else {
          $return = FALSE;
        }
      }

      $return = $return ?? TRUE;
    }
    catch (RequirementsException $e) {
      $return = FALSE;
    }

    unset($clone);
    unset($required_migrations);
    unset($requirements);

    return $return;
  }

  /**
   * Returns the IDs of the given migration's unmet migration dependencies.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance to check.
   */
  public function getMissingRequirements(MigrationInterface $migration) {
    $clone = clone $migration;
    $requirements = $clone->getRequirements();
    if (!empty($requirements)) {
      // Requirements need to be checked.
      $required_migrations = $this->createInstances($requirements);
      $missing_migrations = array_diff($requirements, array_keys($required_migrations));

      // Check if the dependencies are in good shape.
      foreach ($required_migrations as $required_migration) {
        if (!$required_migration->allRowsProcessed()) {
          $missing_migrations[] = $required_migration->id();
        }
      }
    }

    unset($clone);
    unset($requirements);
    unset($required_migrations);
    unset($required_migration);

    return $missing_migrations ?? [];
  }

  /**
   * Returns requirement message of the given plugin, if any.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance to check.
   *
   * @return string|null
   *   The requirements-not-met message if not met, or NULL if the migration is
   *   ready for being executed.
   */
  public function getMigrationRequirementsMessages(MigrationInterface $migration): ?string {
    $message = NULL;
    $migration_clone = clone $migration;
    // Check whether the current migration source and destination plugin
    // requirements are met or not.
    try {
      if ($migration_clone->getSourcePlugin() instanceof RequirementsInterface) {
        $migration_clone->getSourcePlugin()->checkRequirements();
      }
      if ($migration_clone->getDestinationPlugin() instanceof RequirementsInterface) {
        $migration_clone->getDestinationPlugin()->checkRequirements();
      }

      if (!empty($missing_migration_ids = $this->getMissingRequirements($migration_clone))) {
        $message = sprintf("Missing migrations %s.", implode(', ', $missing_migration_ids));
      }
    }
    catch (RequirementsException $e) {
      $message = $e->getMessage();
    }

    unset($migration_clone);
    unset($missing_migration_ids);

    return $message;
  }

}
